/*----------------------------------------------------------------------------

  NLHDR - Non local high dynamic range image generation

  Copyright (c) 2013 cecilia aguerrebere <caguerrebere@gmail.com>
   
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  This code is partially based on the implementation by A. Buades
  
  A. Buades, B. Coll, and J.-M. Morel. Non-Local Means Denoising. 
  Image Processing On Line, 2011

 ----------------------------------------------------------------------------*/
  
#include "mex.h"
#include "libauxiliar.h"
#include <omp.h>

void global_transl_estimations( double ***fpI, int iFrames, int iChannels, int ref_t, int iwxh, int iWidth, int iHeight, double ** prnu, double muR, double* g, double iVsat, double* tau, int y_min, int y_max, int x_min, int x_max, int *trans_y_out, int* trans_x_out) { 
	
  /* Global alignment for correct search window centering */		
  int c,z;
    
  int trans_y;       /* Vector of estimated vertical translations (in rows) of 
				 each image with respect to the reference image */
  int trans_x;       /* Vector of estimated horizontal translations (in columns) 
				 of each image with respect to the reference image */

    /* Allocate memory for the irradiance and variance maps for translation estimation */
    double *rad_movEst[iFrames][iChannels];		
    int *satMap_movEst[iFrames][iChannels];
  
    /* Compute irradiance for the reference frame */
    for ( c = 0; c < iChannels; c++) {
      rad_movEst[ref_t][c] = (double*) malloc(iwxh*sizeof(double));
      satMap_movEst[ref_t][c] = (int*) malloc(iwxh*sizeof(int));
    }
    fnormalize_frame_onlyRad( fpI[ref_t], rad_movEst[ref_t], iWidth, iHeight, prnu, muR, g, tau[ref_t], iChannels );
    fsaturation_maps( fpI[ref_t], satMap_movEst[ref_t], iWidth, iHeight, iChannels, iVsat );
    
    /* Compute irradiance for the rest of the frames */    
    for (z = 0; z < iFrames; z++)      
      if ( z != ref_t ) {
		  printf("Estimating translation for frame %d.\n",z);
	for (c = 0; c < iChannels; c++) {
	  rad_movEst[z][c] = (double*) malloc(iwxh*sizeof(double));
	  satMap_movEst[z][c] = (int*) malloc(iwxh*sizeof(int));
	}    
	fnormalize_frame_onlyRad( fpI[z], rad_movEst[z], iWidth, iHeight, prnu, muR, g, tau[z], iChannels );
	fsaturation_maps( fpI[z], satMap_movEst[z], iWidth, iHeight, iChannels, iVsat );
		  	
	/* Perform translation estimation using red channel */
	estimate_translation( rad_movEst[ref_t][0], rad_movEst[z][0], satMap_movEst[ref_t][0], satMap_movEst[z][0], iWidth, iHeight,y_min, y_max, x_min, x_max, &trans_y, &trans_x  );
	/*global_mov_estimation( rad_movEst[ref_t][0], rad_movEst[z][0], satMap_movEst[ref_t][0], satMap_movEst[z][0], iWidth, iHeight,y_min, y_max, x_min, x_max, &trans_y, &trans_x );*/
	trans_y_out[z] = trans_y;
	trans_x_out[z] = trans_x;	
	
      } else {
	trans_y_out[z] = 0;
	trans_x_out[z] = 0;
      }
      
}


void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
	/* check for proper number of arguments */
	if(nrhs!=16) {
	    mexErrMsgIdAndTxt("MyToolbox:arrayProduct:nrhs",
	                      "Sixteen inputs required.");
	}
	
	/* Variables definition */
	mwSize y_min, y_max, x_min, x_max;
	mwSize iFrames;
	mwSize iChannels;
	int iWidth;
	int iHeight;
   	mwSize iWidth_pgm;
    mwSize iHeight_pgm;
	double* tau;
	mwSize ref_t;
	double ih;
	double muR;
	double sigma2R;
	double* g;
	double iVsat;	
	
	/* Read input data */
    iFrames = mxGetScalar(prhs[1]);
	iChannels = mxGetScalar(prhs[2]);
    iWidth_pgm = mxGetScalar(prhs[3]);
	iHeight_pgm	= mxGetScalar(prhs[4]); 
	tau = (double*)mxGetPr(prhs[5]);     
	ref_t = mxGetScalar(prhs[6]); 
    muR = mxGetScalar(prhs[7]); 
    sigma2R = mxGetScalar(prhs[8]); 
    g = (double*)mxGetPr(prhs[9]);     	          	      
	iVsat = mxGetScalar(prhs[10]); 
    y_min = mxGetScalar(prhs[11]); 
    y_max = mxGetScalar(prhs[12]); 
    x_min = mxGetScalar(prhs[13]); 
    x_max = mxGetScalar(prhs[14]); 
    
    iWidth = (int)iWidth_pgm / 2;
    iHeight = (int)iHeight_pgm / 2;    

	/* Input data */		
	double* input_img = mxGetPr(prhs[0]); 
	int i,ii,x,y;	

	double** fpI[iFrames];
	double* input_img_aux = (double*)malloc(sizeof(double)*iWidth_pgm*iHeight_pgm);	

	for ( i=0; i<iFrames; i++ ) {
		
		/* Copy in input_img_aux frame i from input_img */		
		int l = i*iWidth_pgm*iHeight_pgm;		
		
		for (x=0; x<iWidth_pgm*iHeight_pgm; x++)
			input_img_aux[x] = (double)input_img[l + x];		
				
		/* Allocate memory for fpI */		
		fpI[i] = (double**) malloc(iChannels*sizeof(double*));
		for (ii=0; ii < iChannels; ii++) {
			fpI[i][ii] = (double*) malloc(iHeight*iWidth*sizeof(double)); 
		}	
			
		separate_channels( input_img_aux, fpI[i], iWidth_pgm, iHeight_pgm ); 
	}

	/* PRNU */
	double *prnu[iChannels];
    for (ii=0; ii < iChannels; ii++) 
		 prnu[ii] = (double*) malloc(iWidth*iHeight*sizeof(double)); 
			
	double* prnu_input = mxGetPr(prhs[15]);
	double* prnu_aux = (double*)malloc(sizeof(double)*iWidth_pgm*iHeight_pgm);	 
	for (x=0; x<iWidth_pgm*iHeight_pgm; x++)
		prnu_aux[x] = (double)prnu_input[x];	    	 
    separate_channels( prnu_aux, prnu, iWidth_pgm, iHeight_pgm );

    /* Allocate memory for translation vectors */		
   int* trans_y = (int*) malloc(sizeof(int)*iFrames);       /* Vector of estimated vertical translations (in rows) of 
                                                            each image with respect to the reference image */
   int* trans_x = (int*) malloc(sizeof(int)*iFrames);       /* Vector of estimated horizontal translations (in columns) 
                                                            of each image with respect to the reference image */

    global_transl_estimations( fpI, iFrames, iChannels, ref_t, iWidth*iHeight, iWidth, iHeight, prnu, muR, g, iVsat, tau, y_min, y_max, x_min, x_max, trans_y, trans_x);
        
    for (x=0; x<iFrames; x++) 
        printf("%d, %d\n",trans_y[x],trans_x[x]);
    
    /* create the output vectors */
	plhs[0] = mxCreateDoubleMatrix( iFrames, 1, mxREAL);
    plhs[1] = mxCreateDoubleMatrix( iFrames, 1, mxREAL);
	/* get a pointer to the real data in the output matrix */
	double* trans_y_out = mxGetPr(plhs[0]);
    double* trans_x_out = mxGetPr(plhs[1]);

    for (x=0; x<iFrames; x++) {
        trans_y_out[x] = (double)trans_y[x];   
        trans_x_out[x] = (double)trans_x[x];   
    }

    
}