/*----------------------------------------------------------------------------

  NLHDR - Non local high dynamic range image generation

  Copyright (c) 2013 cecilia aguerrebere <caguerrebere@gmail.com>
   
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  This code is partially based on the implementation by A. Buades
  
  A. Buades, B. Coll, and J.-M. Morel. Non-Local Means Denoising. 
  Image Processing On Line, 2011

 ----------------------------------------------------------------------------*/
  

#include "libauxiliar.h"
#include <omp.h>
#include <float.h>


double fpatch_dist_ch(double *u0,double *u1,double *var0,double *var1,int *satMap0,int *satMap1,int i0,int j0,int i1,int j1,int radius,int width0, int width1, int minPix, double distSatPatchs, int *total_pix_used) {

	int total_pix = 0;
    double dist=0.0f;
    int r,s;
    
    for (s=-radius; s<= radius; s++) {

        int l = (j0+s)*width0 + (i0-radius);
        double *ptr0 = &u0[l];
        double *ptr0_var = &var0[l];
        int *ptr0_satMap = &satMap0[l];

        l = (j1+s)*width1 + (i1-radius);
        double *ptr1 = &u1[l];
        double *ptr1_var = &var1[l];
		int *ptr1_satMap = &satMap1[l];
		
        for (r=-radius; r<=radius; r++,ptr0++,ptr1++,ptr0_var++,ptr1_var++,ptr0_satMap++,ptr1_satMap++) {
			
			/* Only use non saturated pixels to compute the distance */
			if ( *ptr0_satMap == 1 && *ptr1_satMap == 1 ) {
	            double dif = (*ptr0 - *ptr1);	           
	            double var = (*ptr0_var +  *ptr0_var);
	            dist += ((dif*dif) - var) / (var);	            
	            
	            total_pix++; 

	        }
        }

    }

    if ( total_pix < minPix )   /* Check that at least minPix pixels are not saturated in the patch */
			 dist = distSatPatchs;			 	
	
	*total_pix_used = total_pix;

    return dist;
}




double fpatch_dist(double **u0,double **u1,double **var0,double **var1,int **satMap0,int **satMap1,int i0,int j0,int i1,int j1,int radius,int channels, int width0, int width1, int minPix,double distSatPatchs, int *total_pix_used_out) {

    double dif = 0.0f;
    double total_channels = 0.0;
 	int total_pix_used, total_pix_used_aux = 0;
	int ii;
	
    for (ii=0; ii < channels; ii++) {

        dif += fpatch_dist_ch(u0[ii],u1[ii],var0[ii],var1[ii],satMap0[ii],satMap1[ii],i0,j0,i1,j1,radius,width0, width1,minPix,distSatPatchs,&total_pix_used);	
        total_pix_used_aux += total_pix_used;
        total_channels += 1.0;
    }

    /* Mean distance among channels */
	dif = dif / total_channels;	
	
	/* Mean number of pixels used to compute distance bewteen patchs. */
	total_pix_used_aux = total_pix_used_aux / total_channels;
    
    *total_pix_used_out = total_pix_used_aux;
    
    return dif;
}

	
void fnormalize_frame_ch( double *u0, double *rad, double *var,double *radDvar, double *invVar, int iWidth, int iHeight, double *prnu, double muR, double sigma2R, double g, double tau ) {
	
	int x,y;
	
    for (y=0; y < iHeight ; y++)	{  
	
		int l = y*iWidth; 
		double *ptr0 = &u0[l]; 
		double *ptr1 = &prnu[l];		
		double *ptr_rad = &rad[l];  
		double *ptr_var = &var[l];  
		double *ptr_radDvar = &radDvar[l];  
		double *ptr_invVar = &invVar[l];  
		
		for(x=0 ; x < iWidth;  x++, ptr0++, ptr1++, ptr_rad++, ptr_var++,ptr_radDvar++,ptr_invVar++) {
		    double f = (*ptr1) * g * tau;
		    *ptr_rad = ( *ptr0 - muR ) / f; 
		    *ptr_var = ( g*( *ptr0 - muR ) + 2.0*sigma2R ) / (f*f);
		    if (*ptr_var < 0) {
				printf("Warning: Negative variance value. Mapped to minimum possible value.\n");
				*ptr_var = 2.0*sigma2R / (f*f);
			}
		    *ptr_radDvar = (double)*ptr_rad / (double)*ptr_var;
		    *ptr_invVar = 1.0 / (double)*ptr_var;				    		   
		} 
    }
}

void fnormalize_frame( double **u0, double **rad, double **var, double **radDvar, double **invVar, int iWidth, int iHeight, double **prnu, double muR, double sigma2R, double *g, double tau, int channels )  {
	
	int ii;
	
    for (ii=0; ii < channels; ii++) {

        fnormalize_frame_ch(u0[ii], rad[ii], var[ii], radDvar[ii], invVar[ii], iWidth, iHeight, prnu[ii], muR, sigma2R, g[ii], tau);

    }
}


void separate_channels( double *u0, double **u1, int ncol, int nrow ) 
{
	/* We suppose that the number of rows and columns of the input images 
	 * are even. Otherwise, the size of the output images would be different
	 * for the different channels. */
	
	if ( nrow % 2 != 0 ) /* number of rows is not even */
		printf("Warning: the number of rows has been truncated to be even.\n");

	if ( ncol % 2 != 0 ) /* number of columns is not even */
		printf("Warning: the number of columns has been truncated to be even.\n");
	
	int nrow_ch = nrow / 2;	
	int ncol_ch = ncol / 2;	
	int k,x, y = 0;
	
	for ( k=0; k < nrow_ch; k++, y++ ) {
		int l = y*ncol;
		int h = k*ncol_ch;
		
		double *ptr = &u0[l];
		double *ptrR = &u1[0][h];
		double *ptrG1 = &u1[1][h];
		
		for ( x=0; x < ncol_ch; x++, ptr++, ptrR++, ptrG1++ ) {
				*ptrR = *ptr;   /* Fill red channel */
				ptr++;
				*ptrG1 = *ptr;  /* Fill green channel */
		}

		y++;
		l = y*ncol;
		ptr = &u0[l];
		double *ptrG2 = &u1[2][h];
		double *ptrB = &u1[3][h];
		
		for ( x=0; x < ncol_ch; x++, ptr++, ptrG2++, ptrB++ ) {
				*ptrG2 = *ptr;   /* Fill green channel */
				ptr++;
				*ptrB = *ptr;    /* Fill blue channel */
		}

	}	
}

void separate_channels_int( double *u0, int **u1, int ncol, int nrow ) 
{
	/* We suppose that the number of rows and columns of the input images 
	 * are even. Otherwise, the size of the output images would be different
	 * for the different channels. */
	
	if ( nrow % 2 != 0 ) /* number of rows is not even */
		printf("Warning: the number of rows has been truncated to be even.\n");

	if ( ncol % 2 != 0 ) /* number of columns is not even */
		printf("Warning: the number of columns has been truncated to be even.\n");
	
	int nrow_ch = nrow / 2;	
	int ncol_ch = ncol / 2;	
	int k,x,y = 0;
	
	for ( k=0; k < nrow_ch; k++, y++ ) {
		int l = y*ncol;
		int h = k*ncol_ch;
		
		double *ptr = &u0[l];
		int *ptrR = &u1[0][h];
		int *ptrG1 = &u1[1][h];
		
		for ( x=0; x < ncol_ch; x++, ptr++, ptrR++, ptrG1++ ) {
				*ptrR = (int)*ptr;   /* Fill red channel */
				ptr++;
				*ptrG1 = (int)*ptr;  /* Fill green channel */
		}

		y++;
		l = y*ncol;
		ptr = &u0[l];
		int *ptrG2 = &u1[2][h];
		int *ptrB = &u1[3][h];
		
		for ( x=0; x < ncol_ch; x++, ptr++, ptrG2++, ptrB++ ) {
				*ptrG2 = (int)*ptr;   /* Fill green channel */
				ptr++;
				*ptrB = (int)*ptr;    /* Fill blue channel */
		}

	}	
}

void fsaturation_maps_ch( double *u0, int *sat_map, int iWidth, int iHeight, double vsat ) {

	int y,x;
	
    for( y=0; y < iHeight ; y++)	{  
	
		int l = y*iWidth; 
		double *ptr0 = &u0[l]; 
		int *ptr_satMap = &sat_map[l];  
		
		for( x=0 ; x < iWidth;  x++, ptr0++, ptr_satMap++) {
			if ( *ptr0 >= vsat )			
				*ptr_satMap = 0;
			else
				*ptr_satMap = 1; 
		} 
    }
}

void fsaturation_maps( double **u0, int **sat_map, int iWidth, int iHeight, int channels, double vsat )  {

	int ii;
	
    for (ii=0; ii < channels; ii++) {

        fsaturation_maps_ch(u0[ii], sat_map[ii], iWidth, iHeight, vsat);

    }
}

void remake_bayer( double **u1, double *u0, int ncol_ch, int nrow_ch ) 
{
	
	/* u1 is the input 4 channel image */
	/* u0 is the reconstructed Bayer pattern */
	
	/* We suppose that the number of rows and columns of the input images 
	 * are even. Otherwise, the size of the output images would be different
	 * for the different channels. */
		
	int ncol = ncol_ch * 2;	
	int k,x,y = 0;
	
	for ( k=0; k < nrow_ch; k++, y++ ) {
		int l = y*ncol;
		int h = k*ncol_ch;
		
		double *ptr = &u0[l];
		double *ptrR = &u1[0][h];
		double *ptrG1 = &u1[1][h];
		
		for ( x=0; x < ncol_ch; x++, ptr++, ptrR++, ptrG1++ ) {
				*ptr = *ptrR;   /* Fill red channel */
				ptr++;
				*ptr = *ptrG1;  /* Fill green channel */
		}

		y++;
		l = y*ncol;
		ptr = &u0[l];
		double *ptrG2 = &u1[2][h];
		double *ptrB = &u1[3][h];
		
		for ( x=0; x < ncol_ch; x++, ptr++, ptrG2++, ptrB++ ) {
				*ptr = *ptrG2;  /* Fill green channel */
				ptr++;
				*ptr = *ptrB;   /* Fill blue channel */
		}

	}	
}


void estimate_translation( double *img_ref, double *img, int *satMap_ref, int *satMap, int iWidth, int iHeight,int y_min, int y_max,int x_min, int x_max, int* trans_y_out, int* trans_x_out ) {
	
	/* y row index */
	/* x column index */
	double dist_min = DBL_MAX;
	int trans_x, trans_y,x,y,yt,xt;
	
	/* Compute common saturation mask */
	int *satMap_common = (int*) malloc(iWidth*iHeight*sizeof(int));
	for(y=0; y < iHeight ; y++)	{  			
		int l = y*iWidth; 
				
		int *ptr_ref = &satMap_ref[l]; 				
		int *ptr = &satMap[l];
		int *ptr_common = &satMap_common[l];
				
		for(x=0 ; x < iWidth;  x++, ptr++, ptr_ref++, ptr_common++) {	
			*ptr_common = (*ptr_ref) * (*ptr);			 
		}		
	}							
    	
#pragma omp parallel private(yt,xt,y,x) 
	for ( yt=y_min; yt <= y_max; yt+=2 ) 
		for ( xt=x_min; xt <= x_max; xt+=2 )	{
			
			double diff;
			int iHeight_t = iHeight - abs(yt);
			int iWidth_t = iWidth - abs(xt);
			double total_pix = (double) total_valid_pixels( satMap_common, yt, xt, iHeight_t, iWidth_t, iWidth);
						
			/* Define the origin of img depending on the translations */		
			int x_orig_ref, y_orig_ref, x_orig, y_orig; 
			
			( yt < 0 ) ? ( y_orig_ref = 0, y_orig = -yt ) : ( y_orig_ref = yt, y_orig = 0 );
			( xt < 0 ) ? ( x_orig_ref = 0, x_orig = -xt ) : ( x_orig_ref = xt, x_orig = 0 );
						
			/* Scan the images to compute the distance */
			double dist = 0.0;
			int pix_included = 0;
			
		    for(y=0; y < iHeight_t ; y++)	{  			
				int l = (y + y_orig_ref)*iWidth; 
				int lt = (y + y_orig)*iWidth; 
				 
				double *ptr_ref = &img_ref[l + x_orig_ref]; 				
				double *ptr = &img[lt + x_orig];	
				int *ptr_common = &satMap_common[l + x_orig_ref];							
				
				for(x=0 ; x < iWidth_t;  x++, ptr++, ptr_ref++,ptr_common++) {					
					if ( dist < dist_min )	{  /* stop computing the distance if it is already above 
					                              the larger distance */
						if ( *ptr_common == 1 ) { /* if the pixel is not saturated in both images */
							diff = *ptr_ref - *ptr;						
							dist += diff*diff / total_pix;
							pix_included++;
						}														
					} else {  /* finish both for loops */
						x = iWidth_t;
						y = iHeight_t;
					}					
				} 
		    }
		    
		    if ( dist < dist_min && pix_included == total_pix ) {
				#pragma omp critical
				if ( dist < dist_min && pix_included == total_pix ) {
					dist_min = dist;
					trans_x = xt;
					trans_y = yt;
					
				}								
			} 
		    			
		} /* end for xt */
		
	*trans_y_out = trans_y;
	*trans_x_out = trans_x;
		
	
}

void fnormalize_frame_onlyRad_ch( double *u0, double *rad, int iWidth, int iHeight, double *prnu, double muR, double g, double tau ) {

	int x,y;

    for(y=0; y < iHeight ; y++)	{  
	
		int l = y*iWidth; 
		double *ptr0 = &u0[l]; 
		double *ptr1 = &prnu[l];		
		double *ptr_rad = &rad[l];  
		
		for(x=0 ; x < iWidth;  x++, ptr0++, ptr1++, ptr_rad++) {
		    double f = (*ptr1) * g * tau;
		    *ptr_rad = ( *ptr0 - muR ) / f; 
		} 
    }
}

void fnormalize_frame_onlyRad( double **u0, double **rad, int iWidth, int iHeight, double **prnu, double muR, double *g, double tau, int channels )  {

	int ii;
	
    for (ii=0; ii < channels; ii++) {

        fnormalize_frame_onlyRad_ch(u0[ii], rad[ii], iWidth, iHeight, prnu[ii], muR, g[ii], tau);

    }
}


int total_valid_pixels( int *satMap_common, int yt, int xt, int iHeight_t, int iWidth_t, int iWidth) {

	/* Define the origin of img depending on the translations */	
	int x_orig_ref, y_orig_ref, x_orig, y_orig; 
			
	( yt < 0 ) ? ( y_orig_ref = 0, y_orig = -yt ) : ( y_orig_ref = yt, y_orig = 0 );
	( xt < 0 ) ? ( x_orig_ref = 0, x_orig = -xt ) : ( x_orig_ref = xt, x_orig = 0 );
						
	/* Scan the images to compute the total pixels */
	int x,y,pix_included = 0;
	
	   for(y=0; y < iHeight_t ; y++)	{  
		   			
		int l = (y + y_orig_ref)*iWidth; 

		int *ptr_common = &satMap_common[l + x_orig_ref];							
				
		for(x=0 ; x < iWidth_t;  x++, ptr_common++) 					
				if ( *ptr_common == 1 )  /* if the pixel is not saturated in both images */
					pix_included++;
																		
		}
		
		return pix_included; 
	   
}

void error(const char * msg)
{
  fprintf(stderr,"Fatal Error: %s\n",msg);
  exit(EXIT_FAILURE);
}

void print_info ( int win, int bloc, double ih, int ref_t, int iFrames, double* tau, int d_c,	double* g, double muR, double sigma2R, double iVsat, char** input, int* trans_y, int* trans_x, const char* prnu ) {
	
	printf("------------------------------------------------- \n");
	printf("Non local high dynamic range image generation. \n");
	printf("------------------------------------------------- \n");
	printf("\n");
	printf("Parameters setting: \n");
	printf("------------------------------------------------- \n");
	printf("Patch size = %d x %d\n",2*win+1,2*win+1);
	printf("Research window size = %d x %d\n",2*bloc+1,2*bloc+1);
	printf("h = %f\n",ih);
	printf("\n");	
	printf("Camera parameters\n");
	printf("------------------------------------------------- \n");
	printf("Nb frames = %d\n",iFrames);
	printf("Reference frame = %d\n",ref_t);
	printf("Exposures = ");
	int j;
	for (j=0; j<iFrames; j++)
		printf("%f ",tau[j]);
    printf("\n");
    printf("Channels = %d\n",d_c);
    printf("\n");
	printf("Gain = ");
	for (j=0; j<d_c; j++)
		printf("%.4f ",g[j]);
	printf("\n");
	printf("Readout mean = %.1f\n",muR);
	printf("Readout variance = %.1f\n",sigma2R);
	printf("Saturation threshold = %.1f\n",iVsat);
	if ( strcmp(prnu,"0") == 0)
		printf("PRNU file not specified. Factors set to 1.\n");
	else
		printf("PRNU file = %s.\n",prnu);
	printf("\n");
	printf("Input images:\n");
	for (j=0; j<iFrames; j++)
		printf("%s \n",input[j]);		
	printf("\n");
	printf("Translation values ( vertical, horizontal ) \n");
	printf("with respect to the reference frame  \n");
	printf("------------------------------------------------- \n");
	for (j=0; j<iFrames; j++)
		printf("( %d, %d )\n",trans_y[j],trans_x[j]);		
	printf("\n");
	
	
}	
	
	

